# SPDX-License-Identifier: LGPL-3.0-only AND Apache-2.0
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>
# Copyright 2015 The Prometheus Authors
# Portions of this file are adapted from prometheus_client

"""Metric implementations"""

import bisect
from contextlib import contextmanager
import itertools
import sys
import threading
import time

from prometheus_client import metrics, metrics_core, registry, samples

import _mpmetrics
from .atomic import AtomicUInt64, AtomicDouble
from .generics import IntType
from .heap import Heap
from .types import Box, Dict, Double, Array, List, Struct, UInt64
from .util import classproperty, genmask

@contextmanager
def Timer(callback):
    now = time.perf_counter()
    yield
    callback(max(time.perf_counter() - now, 0))

def _validate_labelname(label):
    if not metrics_core.METRIC_LABEL_NAME_RE.match(label):
        raise ValueError(f"invalid label {label}")

    if metrics_core.RESERVED_METRIC_LABEL_NAME_RE.match(label):
        raise ValueError(f"reserved label {label}")

def _validate_exemplar(exemplar):
    code_points = 0
    for k, v in exemplar.items():
        _validate_labelname(k)
        code_points += len(k)
        code_points += len(v)

    if code_points > 128:
        raise ValueError("exemplar too long ({code_points} code points)")

class Collector:
    """A basic collector for non-labeled metrics.

    Attributes are proxied to the underlying metric.
    """
    def __init__(self, metric, name, docs, registry, heap, kwargs):
        self._name = name
        self._docs = docs
        self._metric = metric(heap, **kwargs)
        self.__doc__ = metric.__doc__
        
        registry.register(self)

    def __getattr__(self, name):
        if name == '__getstate__':
            raise AttributeError

        try:
            return getattr(self.__dict__['_metric'], name)
        except KeyError:
            raise AttributeError

    def _family(self):
        return metrics_core.Metric(self._name, self._docs, self._metric._typ)

    def describe(self):
        """Describe the metric

        :return: An iterator yielding one metric with no samples
        :rtype: Iterator[prometheus_client.metrics_core.Metric]
        """

        yield self._family()

    def collect(self):
        """Collect samples from the metric

        :return: An iterator yielding one metric with collected samples.
        :rtype: Iterator[prometheus_client.metrics_core.Metric]
        """

        family = self._family()
        def add_sample(suffix, value, labels={}, exemplar=None):
            family.add_sample(self._name + suffix, labels, value, exemplar=exemplar)
        self._metric._sample(add_sample, self._name)
        yield family

class LabeledCollector(Struct):
    """A collector supporting labeled metrics.

    :py:func:`labels` must be called to get individual metrics.
    """
    _fields_ = {
        '_shared_lock': _mpmetrics.Lock,
        '_metrics': Dict,
    }

    def __init__(self, mem, metric, name, docs, labelnames, registry, kwargs, heap):
        super().__init__(mem, heap=heap)

        self._metric = metric
        self._name = name
        self._docs = docs
        self._kwargs = kwargs
        self._heap = heap
        self.__doc__ = metric.__doc__

        self._labelnames = tuple(labelnames)
        for label in self._labelnames:
            _validate_labelname(label)

            if hasattr(self._metric, '_reserved_labels') and label in self._metric._reserved_labels:
                raise ValueError(f"reserved label {label}")

            if getattr(self._metric, '_name_is_reserved', False) and label == name:
                raise ValueError(f"reserved label {label}")

        self._lock = threading.Lock()
        self._cache = dict()
        registry.register(self)

    def _getstate(self):
        return {
            'metric': self._metric,
            'name': self._name,
            'docs': self._docs,
            'kwargs': self._kwargs,
            'labelnames': self._labelnames,
        }

    def _setstate(self, mem, metric, name, docs, kwargs, labelnames, heap, **others):
        super()._setstate(mem, heap)
        self._metric = metric
        self._name = name
        self._docs = docs
        self._kwargs = kwargs
        self._heap = heap
        self._labelnames = labelnames
        self._lock = threading.Lock()
        self._cache = dict()

    def _label_values(self, values, labels):
        if values and labels:
            raise ValueError("can't pass both *args and **kwargs")

        if labels:
            if sorted(labels) != sorted(self._labelnames):
                raise ValueError("incorrect label names")
            values = (labels[label] for label in self._labelnames)
        else:
            if len(values) != len(self._labelnames):
                raise ValueError("incorrect label count")
        return tuple(sys.intern(str(label)) for label in values)

    def labels(self, *values, **labels):
        """Return the child for the given labelset.

        All metrics can have labels, allowing grouping of related time series.
        Taking a counter as an example::

            from mpmetrics import Counter

            c = Counter('my_requests_total', 'HTTP Failures', ['method', 'endpoint'])
            c.labels('get', '/').inc()
            c.labels('post', '/submit').inc()

        Labels can also be provided as keyword arguments::

            from mpmetrics import Counter

            c = Counter('my_requests_total', 'HTTP Failures', ['method', 'endpoint'])
            c.labels(method='get', endpoint='/').inc()
            c.labels(method='post', endpoint='/submit').inc()

        See the best practices on `naming <http://prometheus.io/docs/practices/naming/>`_
        and `labels <http://prometheus.io/docs/practices/instrumentation/#use-labels>`_.
        """

        values = self._label_values(values, labels)

        with self._lock:
            metric = self._cache.get(values)
            if not metric:
                with self._shared_lock:
                    metric = self._metrics.get(values)
                    if not metric:
                        metric = self._metric(self._heap, **self._kwargs)
                        self._metrics[values] = metric
                self._cache[values] = metric
            return metric

    def _family(self):
        return metrics_core.Metric(self._name, self._docs, self._metric._typ)

    def describe(self):
        """Describe the metric

        :return: An iterator yielding one metric with no samples
        :rtype: Iterator[prometheus_client.metrics_core.Metric]
        """

        yield self._family()

    def collect(self):
        """Collect samples from the metric

        :return: An iterator yielding one metric with samples collected from each label.
        :rtype: Iterator[prometheus_client.metrics_core.Metric]
        """

        family = self._family()
        with self._lock:
            with self._shared_lock:
                for values, metric in self._metrics.items():
                    if values not in self._cache:
                        self._cache[values] = metric
            metrics = self._cache

        for labelvalues, metric in metrics.items():
            metric_labels = dict(zip(self._labelnames, labelvalues))
            def add_sample(suffix, value, labels={}, exemplar=None):
                family.add_sample(self._name + suffix, metric_labels | labels, value,
                                  exemplar=exemplar)
            metric._sample(add_sample, self._name)
        yield family

class CollectorFactory:
    """A factory for creating new metrics.

    This class is used by metrics to create the appropriate collector based on the constructor
    arguments.
    """

    _heap_lock = threading.Lock()

    @classproperty
    def heap(cls):
        with cls._heap_lock:
            if not hasattr(cls, '_heap'):
                cls._heap = Heap()
            return cls._heap

    def __init__(self, metric):
        self._metric = metric
        self.__doc__ = metric.__doc__

    def __getattr__(self, name):
        return getattr(self.__dict__['_metric'], name)

    def __call__(self, name, documentation, labelnames=(), namespace="",
                 subsystem="", unit="", registry=registry.REGISTRY, **kwargs):
        """Create a new metric.

        :param str name: The name of the metric
        :param str documentation: Documentation for the metric. This will be displayed as a ``HELP``
            comment before the metric.
        :param Iterable[str] labelnames: A list of labels to be used with the metric
        :param str namespace: A global namespace for the metric. This will be prepended to `name`.
        :param str subsystem: A subsystem name for the metric. This will be prepended to `name`
            after `namespace`.
        :param str unit: The unit of measurement for the metric. This will be appended to `name`.
        :param prometheus_client.registry.CollectorRegistry registry: The registry to register this
            metric with. It will collect data from the metric.
        :param \\**kwargs: Any additional arguments are passed to the metric itself.
        :return: A new metric
        :rtype: Option[Collector, LabeledCollector]

        The name of the metric is roughly::

            name = f"{namespace}_{subsystem}_{name}_{unit}"

        with unnecessary underscores ommitted.

        If `labelnames` is truthy, then a :py:class:`LabeledCollector` for the metric will be
        returned. Otherwise a :py:class:`Collector` will be returned.
        """

        parts = []
        if namespace:
            parts.append(namespace)
        if subsystem:
            parts.append(subsystem)
        if self._metric._typ == 'counter':
            name = name.removesuffix('_total')
        if unit:
            name = name.removesuffix('_' + unit)
        parts.append(name)
        if unit:
            if self._metric._typ in ('info', 'stateset'):
                raise ValueError(f"{self._metric._typ} metrics cannot have a unit")
            parts.append(unit)

        name = '_'.join(parts)
        if not metrics_core.METRIC_NAME_RE.match(name):
            raise ValueError(f"invalid metric name {name}")

        heap = getattr(registry, 'heap', self.heap)

        if labelnames:
            return Box[LabeledCollector](heap, self._metric, name, documentation, labelnames,
                                         registry, kwargs)
        return Collector(self._metric, name, documentation, registry, heap, kwargs)

class Counter(Struct):
    """A Counter tracks counts of events or running totals.

    Example use cases for Counters:

    * Number of requests processed
    * Number of items that were inserted into a queue
    * Total amount of data that a system has processed

    Counters can only go up (and are reset when the process restarts). If your use case can go down,
    you should use a Gauge instead.

    An example for a Counter::

        from mpmetrics import Counter

        c = Counter('my_failures_total', 'Description of counter')
        c.inc()     # Increment by 1
        c.inc(1.6)  # Increment by given value

    There are also utilities to count exceptions raised::

        @c.count_exceptions()
        def f():
            pass

        with c.count_exceptions():
            pass

        # Count only one type of exception
        with c.count_exceptions(ValueError):
            pass

    For more information about the parameters used when creating a `Counter`, refer to
    :py:class:`~mpmetrics.metrics.CollectorFactory`.
    """

    _typ = 'counter'
    _fields_ = {
        '_lock': _mpmetrics.Lock,
        '_total': AtomicUInt64,
        '_created': Double,
        '_exemplar_amount': UInt64,
        '_exemplar_timestamp': Double,
        '_exemplar_labels': Dict,
    }

    def __init__(self, mem, heap, **kwargs):
        super().__init__(mem, heap)
        self._created.value = time.time()

    def inc(self, amount=1, exemplar=None):
        """Increment by the given amount."""
        if amount < 0:
            raise ValueError("amount must be positive")

        if exemplar is not None:
            _validate_exemplar(exemplar)
   
        self._total.add(amount)
        if exemplar is not None:
            with self._lock:
                self._exemplar_amount.value = amount
                self._exemplar_timestamp.value = time.time()
                self._exemplar_labels.clear()
                self._exemplar_labels |= exemplar

    def _sample(self, add_sample, name):
        with self._lock:
            timestamp = self._exemplar_timestamp.value
            if timestamp:
                amount = self._exemplar_amount.value
                labels = self._exemplar_labels.copy()
        add_sample('_total', self._total.get(),
                   exemplar=samples.Exemplar(labels, amount, timestamp) if timestamp else None)
        add_sample('_created', self._created.value)

    @contextmanager
    def count_exceptions(self, exception=Exception):
        """Count exceptions in a block of code or function.

        Can be used as a function decorator or context manager.
        Increments the counter when an exception of the given
        type is raised up out of the code.
        """

        try:
            yield
        except exception:
            self.inc()

Counter = CollectorFactory(Box[Counter])

class Gauge(Struct):
    """Gauge metric, to report instantaneous values.

    Examples of Gauges include:

    * In-progress requests
    * Number of items in a queue
    * Free memory
    * Total memory
    * Temperature

    Gauges can go both up and down::

       from mpmetrics import Gauge

       g = Gauge('my_inprogress_requests', 'Description of gauge')
       g.inc()      # Increment by 1
       g.dec(10)    # Decrement by given value
       g.set(4.2)   # Set to a given value

    There are utilities for common use cases::

       g.set_to_current_time()   # Set to current unix time

       # Increment when entered, decrement when exited.
       @g.track_inprogress()
       def f():
           pass

       with g.track_inprogress():
           pass

    A Gauge can also take its value from a callback::

       d = Gauge('data_objects', 'Number of objects')
       my_dict = {}
       d.set_function(lambda: len(my_dict))

    For more information about the parameters used when creating a `Gauge`, refer to
    :py:class:`~mpmetrics.metrics.CollectorFactory`.
    """

    _typ = 'gauge'
    _fields_ = {
        '_value': AtomicDouble,
    }

    def __init__(self, mem, **kwargs):
        super().__init__(mem)

    def inc(self, amount=1):
        """Increment by the given amount."""
        self._value.add(amount)

    def dec(self, amount=1):
        """Decrement by the given amount."""
        self._value.add(-amount)

    def set(self, amount):
        """Set to the given amount."""
        self._value.set(amount)

    def _sample(self, add_sample, name):
        add_sample('', self._value.get())

    def set_to_current_time(self):
        """Set to the current time in seconds since the Epoch."""
        self.set(time.time())

    @contextmanager
    def track_inprogress(self):
        """Track in-progress blocks of code or functions.

        Can be used as a function decorator or context manager.
        Increments the gauge when the code is entered,
        and decrements when it is exited.
        """

        self.inc()
        try:
            yield
        finally:
            self.dec()

    def time(self):
        """Time a block of code or function, and set the duration in seconds.

        Can be used as a function decorator or context manager.
        """

        return Timer(self.set)

Gauge = CollectorFactory(Box[Gauge])

class _SummaryData(Struct):
    _fields_ = {
        'sum': AtomicDouble,
        'count': AtomicUInt64,
    }

class Summary(Struct):
    """A Summary tracks the size and number of events.

    Example use cases for Summaries:

    * Response latency
    * Request size

    Example for a Summary::

        from mpmetrics import Summary

        s = Summary('request_size_bytes', 'Request size (bytes)')
        s.observe(512)  # Observe 512 (bytes)

    Example for a Summary using time::

        from mpmetrics import Summary

        REQUEST_TIME = Summary('response_latency_seconds', 'Response latency (seconds)')

        @REQUEST_TIME.time()
        def create_response(request):
          '''A dummy function'''
          time.sleep(1)

    Example for using the same Summary object as a context manager::

        with REQUEST_TIME.time():
            pass  # Logic to be timed

    For more information about the parameters used when creating a `Summary`, refer to
    :py:class:`~mpmetrics.metrics.CollectorFactory`.
    """

    _typ = 'summary'
    _reserved_labels = ('quantile',)
    _fields_ = {
        '_lock': _mpmetrics.Lock,
        '_data': Array[_SummaryData, 2],
        '_count': AtomicUInt64,
        '_created': Double,
    }

    def __init__(self, mem, **kwargs):
        super().__init__(mem)
        self._created.value = time.time()

    def observe(self, amount):
        """Observe the given amount.

        The amount is usually positive or zero. Negative values are
        accepted but prevent current versions of Prometheus from
        properly detecting counter resets in the sum of
        observations. See
        https://prometheus.io/docs/practices/histograms/#count-and-sum-of-observations
        for details.
        """

        data = self._data[self._count.add(1) >> 63]
        data.sum.add(amount)
        data.count.add(1)

    def _sample(self, add_sample, name):
        with self._lock:
            count = self._count.add(1 << 63, raise_on_overflow=False)
            hot = self._data[~count >> 63]
            cold = self._data[count >> 63]
            count &= genmask(62, 0)

            while cold.count.get() != count:
                time.sleep(0)

            sum = cold.sum.get()
            hot.count.add(count)
            hot.sum.add(sum)
            cold.count.set(0)
            cold.sum.set(0)

        add_sample('_count', count)
        add_sample('_sum', sum)
        add_sample('_created', self._created.value)

    def time(self):
        """Time a block of code or function, and observe the duration in seconds.

        Can be used as a function decorator or context manager.
        """

        return Timer(self.observe)

Summary = CollectorFactory(Box[Summary])

def _HistogramData(__name__, bucket_count):
    _fields_ = {
        'buckets': Array[AtomicUInt64, bucket_count],
        'sum': AtomicDouble,
        'count': AtomicUInt64,
    }

    ns = locals()
    del ns['bucket_count']
    return type(__name__, (Struct,), ns)

_HistogramData = IntType('_HistogramData', _HistogramData)

def _Histogram(__name__, bucket_count):
    """A Histogram tracks the size and number of events in buckets.

    You can use Histograms for aggregatable calculation of quantiles.

    Example use cases:

    * Response latency
    * Request size

    Example for a Histogram::

        from mpmetrics import Histogram

        h = Histogram('request_size_bytes', 'Request size (bytes)')
        h.observe(512)  # Observe 512 (bytes)

    Example for a Histogram using time::

        from mpmetrics import Histogram

        REQUEST_TIME = Histogram('response_latency_seconds', 'Response latency (seconds)')

        @REQUEST_TIME.time()
        def create_response(request):
          '''A dummy function'''
          time.sleep(1)

    Example of using the same Histogram object as a context manager::

        with REQUEST_TIME.time():
            pass  # Logic to be timed

    For more information about the parameters used when creating a `Summary`, refer to
    :py:class:`~mpmetrics.metrics.CollectorFactory`.

    The default buckets are intended to cover a typical web/rpc request from milliseconds to
    seconds. They can be overridden by passing `buckets` keyword argument to `Histogram`.
    """

    _typ = 'histogram'
    _fields_ = {
        '_thresholds': Array[Double, bucket_count],
        '_lock': _mpmetrics.Lock,
        '_data': Array[_HistogramData[bucket_count], 2],
        '_count': AtomicUInt64,
        '_created': Double,
        '_exemplars': List,
    }

    def __init__(self, mem, thresholds, heap, **kwargs):
        Struct.__init__(self, mem, heap)
        assert len(thresholds) == bucket_count
        self.thresholds = thresholds
        for threshold, initial in zip(self._thresholds, thresholds):
            threshold.value = initial
            self._exemplars.append(None)
        self._created.value = time.time()

    def _setstate(self, mem, heap):
        Struct._setstate(self, mem, heap)
        self.thresholds = tuple(threshold.value for threshold in self._thresholds)

    def observe(self, amount, exemplar=None):
        """Observe the given amount.

        The amount is usually positive or zero. Negative values are
        accepted but prevent current versions of Prometheus from
        properly detecting counter resets in the sum of
        observations. See
        https://prometheus.io/docs/practices/histograms/#count-and-sum-of-observations
        for details.
        """

        if exemplar is not None:
            _validate_exemplar(exemplar)

        i = bisect.bisect_left(self.thresholds, amount)

        data = self._data[self._count.add(1) >> 63]
        data.buckets[i].add(1)
        data.sum.add(amount)
        data.count.add(1)
        if exemplar:
            with self._lock:
                self._exemplars[i] = (exemplar, amount, time.time())

    def _sample(self, add_sample, name):
        with self._lock:
            count = self._count.add(1 << 63, raise_on_overflow=False)
            hot = self._data[~count >> 63]
            cold = self._data[count >> 63]
            count &= genmask(62, 0)

            while cold.count.get() != count:
                time.sleep(0)

            buckets = [bucket.get() for bucket in cold.buckets]
            sum = cold.sum.get()
            exemplars = list(self._exemplars)

            for i, bucket in enumerate(buckets):
                hot.buckets[i].add(bucket)
            hot.sum.add(sum)
            hot.count.add(count)

            for bucket in cold.buckets:
                bucket.set(0)
            cold.sum.set(0)
            cold.count.set(0)

        for val, le, exemplar in zip(itertools.accumulate(buckets), self.thresholds, exemplars):
            add_sample('_bucket', val, { 'le': str(le) },
                       samples.Exemplar(*exemplar) if exemplar else None)
        add_sample('_sum', sum)
        add_sample('_count', count)
        add_sample('_created', self._created.value)

    def _time(self):
        """Time a block of code or function, and observe the duration in seconds.

        Can be used as a function decorator or context manager.
        """

        return Timer(self.observe)

    ns = locals()
    ns['__doc__'] = __doc__
    ns['time'] = ns['_time']
    del ns['_time']
    del ns['bucket_count']

    return type(__name__, (Struct,), ns)

_Histogram = IntType('_Histogram', _Histogram)

class _HistogramFactory:
    __doc__ = _Histogram.cls.__doc__
    _typ = 'histogram'
    _reserved_labels = ('le',)
    DEFAULT_BUCKETS = (.005, .01, .025, .05, .075, .1, .25, .5, .75, 1.0, 2.5, 5.0, 7.5, 10.0,
                       float('inf'))

    def __getattr__(self, name):
        return getattr(_Histogram[len(self.DEFAULT_BUCKETS)], name)

    def __call__(self, heap, buckets=DEFAULT_BUCKETS, **kwargs):
        thresholds = [float(b) for b in buckets]
        if thresholds != sorted(thresholds):
            raise ValueError('thresholds not in sorted order')
        if thresholds and thresholds[-1] != float('inf'):
            thresholds.append(float('inf'))
        if len(thresholds) < 2:
            raise ValueError('must have at least two thresholds')
        thresholds = tuple(thresholds)

        histogram = Box[_Histogram[len(thresholds)]]
        return histogram(heap, thresholds=thresholds, **kwargs)

Histogram = CollectorFactory(_HistogramFactory())

class Enum(Struct):
    """Enum metric, which has one selected state in a set

    Example usage::

        from mpmetrics import Enum

        e = Enum('task_state', 'Description of enum',
                 states=['starting', 'running', 'stopped'])
        e.state('running')

    The first listed state will be the default.

    For more information about the parameters used when creating a `Summary`, refer to
    :py:class:`~mpmetrics.metrics.CollectorFactory`.
    """

    _typ = 'stateset'
    _name_is_reserved = True
    _fields_ = {
        '_value': AtomicUInt64,
    }

    def __init__(self, mem, states, **kwargs):
        super().__init__(mem)
        self._states = states

    def _getstate(self):
        return { 'states': self._states }

    def _setstate(self, mem, heap, states):
        super()._setstate(mem, heap)
        self._states = states

    def state(self, state):
        """Select a given state."""
        self._value.set(self._states.index(state))

    def _sample(self, add_sample, name):
        val = self._value.get()
        for i, state in enumerate(self._states):
            add_sample('', int(i == val), { name: state })

Enum = CollectorFactory(Box[Enum])



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>
# Copyright (c) 2017 Viktor Adam
# Portions of this file are adapted from prometheus_flask_exporter

import time
import random

from flask import Flask
from mpmetrics.flask import PrometheusMetrics

app = Flask(__name__)
PrometheusMetrics(app, group_by="endpoint")

@app.route("/one")
def first_route():
    time.sleep(random.random() * 0.2)
    return "ok"

@app.route("/two")
def the_second():
    time.sleep(random.random() * 0.4)
    return "ok"

@app.route("/three")
def test_3rd():
    time.sleep(random.random() * 0.6)
    return "ok"

@app.route("/four")
def fourth_one():
    time.sleep(random.random() * 0.8)
    return "ok"

@app.route("/error")
def oops():
    return ":(", 500

if __name__ == "__main__":
    app.run("0.0.0.0", 8000, threaded=True)



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

"""
multiprocess-safe atomics

This module contains atomic types which automatically fall back to locking
implementations on architectures which only support 32-bit atomics.

.. py:data:: AtomicDouble

    Either :py:class:`_mpmetrics.AtomicDouble`, or :py:class:`LockingDouble` if
    the former is not supported.

.. py:data:: AtomicInt64

    Either :py:class:`_mpmetrics.AtomicInt64`, or :py:class:`LockingInt64` if
    the former is not supported.

.. py:data:: AtomicUInt64

    Either :py:class:`_mpmetrics.AtomicUInt64`, or :py:class:`LockingUInt64` if
    the former is not supported.
"""

import _mpmetrics
from .types import Double, Int64, UInt64, Struct

# TODO: Rewrite this in C if anyone cares about performance on arches without 64-bit atomics?

class _Locking(Struct):
    _fields_ = {
        '_lock': _mpmetrics.Lock,
    }

    def get(self):
        """Return the current value of the backing atomic"""
        with self._lock:
            return self._value.value

    def set(self, value):
        """Set the backing atomic to `value`."""
        with self._lock:
            self._value.value = value

    def add(self, amount, raise_on_overflow=True):
        """Add 'amount' to the backing atomic.

        :param Union[int, float] amount: The amount to add
        :param bool raise_on_overflow: Whether to raise an exception on overflow
        :return: The value from before the addition.
        :rtype: Union[int, float]
        """

        with self._lock:
            old = self._value.value
            self._value.value = old + amount
            if raise_on_overflow and self._value.value != old + amount:
                raise OverflowError(f"{old} + {amount} too large to fit")
            return old

class LockingDouble(_Locking):
    """An atomic double implemented using a lock"""

    _fields_ = _Locking._fields_ | {
        '_value': Double,
    }

class LockingInt64(_Locking):
    """An atomic 64-bit signed integer implemented using a lock"""

    _fields_ = _Locking._fields_ | {
        '_value': Int64,
    }

class LockingUInt64(_Locking):
    """An atomic 64-bit unsigned integer implemented using a lock"""

    _fields_ = _Locking._fields_ | {
        '_value': UInt64,
    }

AtomicDouble = _mpmetrics.AtomicDouble or LockingDouble
AtomicInt64 = _mpmetrics.AtomicInt64 or LockingInt64
AtomicUInt64 = _mpmetrics.AtomicUInt64 or LockingUInt64



# SPDX-License-Identifier: GPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

from contextlib import nullcontext
import math

from hypothesis import given, strategies as st
import pytest

from mpmetrics.types import Box, Double
from mpmetrics.atomic import AtomicInt64, AtomicUInt64, AtomicDouble

from .common import heap, parallel, parallels, ParallelLoop

@pytest.fixture(scope='module', params=(AtomicInt64, AtomicUInt64, AtomicDouble))
def atomic(request):
    return Box[request.param]

@pytest.fixture(scope='module', params=(AtomicInt64, AtomicUInt64))
def integer(request):
    return Box[request.param]

@given(st.integers())
def test_iset(heap, integer, x):
    a = integer(heap)
    if x not in range(a.min, a.max + 1):
        ctx = pytest.raises(OverflowError)
    else:
        ctx = nullcontext()
    with ctx:
        a.set(x)

def integers():
    return st.integers(AtomicInt64.min, AtomicInt64.max)

@given(integers(), integers())
def test_iadd(heap, x, y):
    a = Box[AtomicInt64](heap)
    a.set(x)
    if x + y not in range(a.min, a.max + 1):
        with pytest.raises(OverflowError):
            a.add(y)
    else:
        a.add(y)
        assert a.get() == x + y

def unsigned_integers():
    return st.integers(AtomicUInt64.min, AtomicUInt64.max)

@given(unsigned_integers(), unsigned_integers())
def test_uadd(heap, x, y):
    a = Box[AtomicUInt64](heap)
    a.set(x)
    if x + y not in range(a.min, a.max + 1):
        with pytest.raises(OverflowError):
            a.add(y)
    else:
        a.add(y)
        assert a.get() == x + y

@given(x=st.floats())
def test_dset(heap, x):
    a = Box[AtomicDouble](heap)
    a.set(x)

@given(x=st.floats(), y=st.floats())
def test_dadd(heap, x, y):
    a = Box[AtomicDouble](heap)
    a.set(x)
    a.add(y)
    if math.isnan(x + y):
        assert math.isnan(a.get())
    else:
        assert a.get() == x + y

class OrderingTest(ParallelLoop):
    def __init__(self, heap, atomic, parallel):
        super().__init__(parallel)
        self.x = atomic(heap)
        self.y = atomic(heap)

    def loop(self, n):
        self.x.add(1)
        self.y.add(1)

    def check(self):
        y = self.y.get()
        x = self.x.get()
        assert x >= y

    def final(self):
        assert self.x.get() == self.total
        assert self.y.get() == self.total

def test_ordering(heap, atomic, parallel):
    OrderingTest(heap, atomic, parallel).run()

class RacyTest(ParallelLoop):
    def __init__(self, heap, parallel):
        super().__init__(parallel)
        self.value = Box[Double](heap)

    def loop(self, n):
        self.value.value += 1

    def final(self):
        assert self.value.value != self.total

def test_racy(heap, parallel):
    if parallel == parallels['thread']:
        pytest.xfail("The GIL isn't racy")
    RacyTest(heap, parallel).run()



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

"""Helpers for pickling polymorphic classes."""

import importlib
import itertools

def saveattr(get):
    """Save the result of `__getattr__`.

    :param get: A `__getattr__` implementation

    Wrap `get` and save the result with `setattr`.
    """
    def wrapped(self, name):
        attr = get(self, name)
        setattr(self, name, attr)
        return attr
    return wrapped

class ObjectType:
    """Helper for classes polymorphic over classes.

    This is a helper class to allow classes which are polymorphic over python
    classes to be pickled. For example::

        import pickle
        from mpmetrics.generics import ObjectType

        def MyClass(__name__, cls):
            return type(__name__, (), locals())

        MyClass = ObjectType('MyClass', MyClass)
        assert MyClass[int].cls is int
        assert pickle.loads(pickle.dumps(MyClass[int]())).cls is int
    """

    class Attr:
        def __init__(self, name, cls, obj, nesting=1):
            self.name = name
            self.cls = cls
            self.obj = obj
            self.nesting = nesting

        @saveattr
        def __getattr__(self, name):
            nesting = self.nesting + (name == '<') - (name == '>')
            if name == '>' and not nesting:
                return self.cls(self.name + '.' + self.obj.__qualname__ + '.>', self.obj)
            else:
                return type(self)(self.name, self.cls, getattr(self.obj, name), nesting)

    class Module:
        def __init__(self, name, cls, parent=None):
            self.name = name
            self.cls = cls
            self.parent = parent

        @saveattr
        def __getattr__(self, name):
            try:
                if self.parent:
                    module = self.parent.__name__ + '.' + name
                else:
                    module = name
                return type(self)(self.name, self.cls, importlib.import_module(module))
            except ModuleNotFoundError:
                if self.parent:
                    prefix = self.name + '.' + self.parent.__name__
                else:
                    prefix = self.name
                return ObjectType.Attr(prefix, self.cls, getattr(self.parent, name))

    def __init__(self, name, cls):
        self.__qualname__ = name
        self.__doc__ = cls.__doc__
        setattr(self, '<', self.Module(name + '.<', cls))

    def __getitem__(self, cls):
        parent = getattr(self, '<')
        for subpath in itertools.chain(cls.__module__.split('.'), cls.__qualname__.split('.')):
            parent = getattr(parent, subpath)
        return getattr(parent, '>')

class IntType:
    """Helper for classes polymorphic over integers.

    This is a helper class to allow classes which are polymorphic over ints
    to be pickled. For example::

        import pickle
        from mpmetrics.generics import IntType

        def MyClass(__name__, x):
            return type(__name__, (), locals())

        MyClass = IntType('MyClass', MyClass)
        assert MyClass[5].x == 5
        assert pickle.loads(pickle.dumps(MyClass[5]())).x == 5
    """

    def __init__(self, name, cls):
        self.__qualname__ = name
        self.__doc__ = cls.__doc__
        self.name = name
        self.cls = cls

    @saveattr
    def __getattr__(self, attr):
        return self.cls(self.name + '.' + attr, int(attr))

    def __getitem__(self, n):
        return getattr(self, repr(n))

class FloatType:
    """Helper for classes polymorphic over floats.

    This is a helper class to allow classes which are polymorphic over floats
    to be pickled. For example::

        import pickle
        from mpmetrics.generics import FloatType

        def MyClass(__name__, x):
            return type(__name__, (), locals())

        MyClass = FloatType('MyClass', MyClass)
        assert MyClass[2.7].x == 2.7
        assert pickle.loads(pickle.dumps(MyClass[2.7]())).x == 2.7
    """

    def __init__(self, name, cls):
        self.__qualname__ = name
        self.__doc__ = cls.__doc__
        self.name = name
        self.cls = cls

    @saveattr
    def __getattr__(self, attr):
        return self.cls(self.name + '.' + attr, float(attr.replace('_', '.')))

    def __getitem__(self, n):
        return getattr(self, repr(n).replace('.', '_'))


class ProductType:
    """Helper to combine other types.

    This is a helper class to allow classes which are polymorphic over multiple
    types to be pickled. For example::

        import pickle
        from mpmetrics.generics import IntType, ObjectType, ProductType

        def MyClass(__name__, cls, x):
            return type(__name__, (), locals())

        MyClass = ProductType('MyClass', MyClass, (ObjectType, IntType))
        assert MyClass[int, 5].cls is int
        assert MyClass[int, 5].x == 5
        assert pickle.loads(pickle.dumps(MyClass[int, 5]())).x == 5
    """

    def __init__(self, name, cls, argtypes, args=()):
        self.__qualname__ = name
        self.__doc__ = cls.__doc__
        self.name = name
        self.cls = cls
        self.argtype = argtypes[0](self.name, self._chain)
        self.argtypes = argtypes[1:]
        self.args = args

    def _chain(self, name, arg):
        if self.argtypes:
            return type(self)(name, self.cls, self.argtypes, (*self.args, arg))
        return self.cls(name, *self.args, arg)

    @saveattr
    def __getattr__(self, name):
        return getattr(self.argtype, name)

    def __getitem__(self, args):
        if not isinstance(args, tuple):
            args = (args,)

        argtype = self.argtype
        for arg in args:
            argtype = argtype[arg]
        return argtype

class ListType:
    """Helper to combine other types.

    This is a helper class to allow classes which are polymorphic over multiple
    types to be pickled. For example::

        import pickle
        from mpmetrics.generics import IntType, ListType

        def MyClass(__name__, xs):
            return type(__name__, (), locals())

        MyClass = ListType('MyClass', MyClass, IntType)
        assert MyClass[1, 2, 3].xs == (1, 2, 3)
        assert pickle.loads(pickle.dumps(MyClass[1, 2, 3]())).xs == (1, 2, 3)
    """

    def __init__(self, name, cls, elemtype):
        self.__qualname__ = name
        self.__doc__ = cls.__doc__
        self.name = name
        self.cls = cls
        self.elemtype = elemtype
        self.length = IntType(name, self._begin)

    def _begin(self, name, length):
        return ProductType(name, self._end, (self.elemtype,) * length)

    def _end(self, name, *elems):
        return self.cls(name, elems)

    @saveattr
    def __getattr__(self, name):
        return getattr(self.length, name)

    def __getitem__(self, args):
        if not isinstance(args, tuple):
            args = (args,)

        return getattr(self, str(len(args)))[args]



# SPDX-License-Identifier: GPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

from copy import copy
import errno
import time

import pytest

from mpmetrics.types import Box
from _mpmetrics import Lock

from .common import heap, parallel

def test_basics(heap):
    l1 = Box[Lock](heap)

    assert l1.acquire()
    with pytest.raises(OSError):
        l1.acquire()

    l2 = copy(l1)
    with pytest.raises(OSError):
        l2.acquire()

    l1.release()
    with pytest.raises(PermissionError):
        l1.release()

    with l1:
        with pytest.raises(OSError):
            l2.acquire()

    assert l1.acquire(block=False)
    l1.release()

    try:
        assert l1.acquire(timeout=1)
        l1.release()
    except OSError as e:
        if e.errno != errno.ENOTSUP:
            raise

def hold(l, b):
    with l:
        b.wait()
        b.wait()

def test_acquire(heap, parallel):
    l = Box[Lock](heap)
    b = parallel.barrier(2)
    p = parallel.spawn(target=hold, args=(l, b))
    p.start()

    try:
        b.wait()

        assert not l.acquire(block=False)
        now = time.time()
        try:
            assert not l.acquire(timeout=0.001)
        except OSError as e:
            if e.errno != errno.ENOTSUP:
                raise
        else:
            assert now + 0.001 <= time.time()

        b.wait()
        with l:
            pass
    except:
        b.abort()
        raise
    finally:
        p.join()



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2021-22 Sean Anderson <seanga2@gmail.com>

"""A shared memory allocator."""

import itertools
import mmap
from multiprocessing.reduction import DupFd
import os
from tempfile import TemporaryFile
import threading

import _mpmetrics
from .types import Size_t, Struct
from .util import align, _align_check

SC_LEVEL1_DCACHE_LINESIZE = 190
try:
    CACHELINESIZE = os.sysconf(SC_LEVEL1_DCACHE_LINESIZE)
except OSError:
    CACHELINESIZE = 64 # Assume 64-byte cache lines

PAGESIZE = 4096

class Heap(Struct):
    """A shared memory allocator.

    This is a basic arena-style allocator. The core algorithm is (effectively)::

        def malloc(size):
            old_base = base
            base += size
            return old_base

    We do not keep track of free blocks, so :py:meth:`Heap.Block.free` is a no-op.

    Memory is requested from the OS in page-sized blocks. As we don't map all
    of our memory up front, it's possible that different processes will map new
    pages at different addresses. Therefore, we keep track of the address where
    each page is mapped, and ensure blocks do not cross page boundaries.
    Larger-than-page-size blocks are supported by aligning the block to the
    page size and mapping all pages in that block in one go.
    """

    _fields_ = {
        '_shared_lock': _mpmetrics.Lock,
        '_base': Size_t,
    }

    def __init__(self, map_size=PAGESIZE):
        """Create a new Heap.

        :param int map_size: The granularity to use when requesting memory from the OS
        """

        if map_size % mmap.ALLOCATIONGRANULARITY:
            raise ValueError("size must be a multiple of {}".format(mmap.ALLOCATIONGRANULARITY))
        _align_check(map_size)
        self.map_size = map_size

        # File backing our shared memory
        self._file = TemporaryFile()
        self._fd = self._file.fileno()
        # Allocate a page to start with
        os.truncate(self._fd, map_size)

        # Process-local shared memory maps
        self._maps = [mmap.mmap(self._fd, map_size)]
        # Lock for _maps
        self._lock = threading.Lock()

        super().__init__(memoryview(self._maps[0])[:self.size])
        self._base.value = self.size

    def __getstate__(self):
        return self.map_size, DupFd(self._fd)

    def __setstate__(self, state):
        self.map_size, df = state
        self._fd = df.detach()
        self._file = open(self._fd, 'a+b')

        # Process-local shared memory maps
        self._maps = [mmap.mmap(self._fd, self.map_size)]
        # Lock for _maps
        self._lock = threading.Lock()

        super()._setstate(memoryview(self._maps[0])[:self.size])

    class Block:
        """A block of memory allocated from a Heap."""

        def __init__(self, heap, start, size):
            """Create a new Block.

            :param Heap heap: The heap this block is from
            :param int start: The offset of this block within the heap
            :param int size: The size of this block
            """

            self.heap = heap
            self.start = start
            self.size = size

        def deref(self):
            """Dereference this block

            :return: The memory referenced by this block
            :rtype: memoryview

            Dereference the block, faulting in unmapped pages as necessary.
            """

            heap = self.heap
            first_page = int(self.start / heap.map_size)
            last_page = int((self.start + self.size - 1) / heap.map_size)
            nr_pages = last_page - first_page + 1
            page_off = first_page * heap.map_size
            off = self.start - page_off
            with heap._lock:
                if len(heap._maps) <= last_page:
                    heap._maps.extend(itertools.repeat(None, last_page - len(heap._maps) + 1))
                if not self.heap._maps[first_page]:
                    heap._maps[first_page] = mmap.mmap(heap._fd, heap.map_size * nr_pages,
                                                       offset=page_off)
                map = heap._maps[first_page]

            return memoryview(map)[off:off+self.size]

        def free(self):
            """Free this block"""
            pass

    def malloc(self, size, alignment=CACHELINESIZE):
        """Allocate shared memory.

        :param int size: The amount of shared memory to allocate, in bytes
        :param int alignment: The minimum alignment of the memory
        :return: A block of shared memory
        :rtype: Block

        Allocate at least `size` bytes of shared memory. It will be aligned to
        at least `alignment`.
        """

        if size <= 0:
            raise ValueError("size must be strictly positive")
        elif size > self.map_size:
            size = align(size, self.map_size)
        _align_check(alignment)

        with self._shared_lock:
            total = align(self._base.value, self.map_size)
            self._base.value = align(self._base.value, alignment)
            if self._base.value + size >= total:
                os.ftruncate(self._fd, align(total + size, self.map_size))
                self._base.value = total
            start = self._base.value
            self._base.value += size

        return self.Block(self, start, size)



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

"""Various small utilities."""

def _align_mask(x, mask):
    return (x + mask) & ~mask

def _align_check(a):
    if not a or a & (a - 1):
        raise ValueError("{} is not a power of 2".format(a))

def align(x, a):
    """Align `x` to `a`

    :param int x: The value to align
    :param int a: The alignment; must be a power of two
    :return: The smallest multiple of `a` greater than `x`
    :rtype: int
    """

    _align_check(a)
    return _align_mask(x, a - 1)

def align_down(x, a):
    """Align `x` down to `a`

    :param int x: The value to align
    :param int a: The alignment; must be a power of two
    :return: The largest multiple of `a` less than `x`
    :rtype: int
    """

    _align_check(a)
    return x & ~(a - 1)

def genmask(hi, lo):
    """Generate a mask with bits between `hi` `lo` set.

    :param int hi: The highest bit to set, inclusive
    :param int lo: The lowest bit to set, inclusive
    :return: The bitmask
    :rtype: int

    `hi` must be greater than `lo`. Bits are numbered in "little-endian" order,
    starting from zero. The following invariant holds::

        mask = 0
        for n in range(lo, hi):
            mask |= 1 << n
        assert mask == genmask(hi, lo)
    """

    return (-1 << lo) & ~(-1 << (hi + 1))

# From https://stackoverflow.com/a/7864317/5086505
class classproperty(property):
    """Like `property` but for classes."""
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()



# SPDX-License-Identifier: GPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

import pickle

from hypothesis import assume, given, reject, settings, strategies as st
from hypothesis.stateful import Bundle, multiple, RuleBasedStateMachine, rule
import pytest

from mpmetrics.atomic import AtomicInt64, AtomicUInt64, AtomicDouble
from mpmetrics.generics import ObjectType, ListType
from mpmetrics.heap import PAGESIZE, Heap
from mpmetrics.types import Array, Box, Dict, Double, List, Size_t, Struct
from _mpmetrics import Lock

from .common import heap

def GenericStruct(__name__, classes):
    ns = { '_fields_': { str(i): cls for i, cls in enumerate(classes)}}
    return type(__name__, (Struct,), ns)

GenericStruct = ListType('GenericStruct', GenericStruct, ObjectType) 

def recursive_types(classes):
    @st.composite
    def arrays(draw):
        array = Array[draw(classes), draw(st.integers(1, 8))]
        # reject too-large types as soon as we can
        if array.size >= PAGESIZE:
            reject()
        return array

    @st.composite
    def structs(draw):
        n = draw(st.integers(1, 8))
        struct = GenericStruct[draw(st.tuples(*((classes,) * n)))]
        # ditto
        if struct.size >= PAGESIZE:
            reject()
        return struct

    return arrays() | structs()

base_types = st.sampled_from((AtomicInt64, AtomicUInt64, AtomicDouble, Double, Size_t, Lock))
types = st.recursive(base_types, recursive_types)

@given(types)
@settings(max_examples=25)
def test_pickle(heap, cls):
    pickle.loads(pickle.dumps(cls))

    assume(cls.size <= heap.map_size)
    v = Box[cls](heap)
    pickle.loads(pickle.dumps(v))

# We use Size_t because drawing from types is slow
@given(st.integers(min_value=1))
def test_array(heap, n):
    A = Array[Size_t, n]
    assert A.size >= n * Size_t.size

    assume(A.size <= heap.map_size)
    a = Box[A](heap)
    assert len(a) == n

@given(st.integers(max_value=0))
def test_bad_size(n):
    with pytest.raises(ValueError):
        Array[Size_t, n]

@settings(max_examples=25, deadline=5000)
class DictComparison(RuleBasedStateMachine):
    keys = Bundle('keys')
    values = Bundle('values')
    heap = Heap()

    def __init__(self):
        super().__init__()
        self.dict = Box[Dict](self.heap)
        self.model = {}

    @rule(target=keys, k=st.text())
    def key(self, k):
        return k

    @rule(target=values, v=st.text())
    def value(self, v):
        return v

    @rule()
    def len(self):
        assert len(self.model) == len(self.dict)

    @rule(k=keys)
    def getitem(self, k):
        try:
            model_k = self.model[k]
        except KeyError:
            with pytest.raises(KeyError):
                dict_k = self.dict[k]
        else:
            dict_k = self.dict[k]
            assert model_k == dict_k

    @rule(k=keys, v=values)
    def setitem(self, k, v):
        self.model[k] = v
        self.dict[k] = v

    @rule(k=keys)
    def delitem(self, k):
        try:
            del self.model[k]
        except KeyError:
            with pytest.raises(KeyError):
                del self.dict[k]
        else:
            del self.dict[k]

    @rule()
    def iters(self):
        for f in (
            iter,
            reversed,
            lambda d: d.items(),
            lambda d: d.keys(),
            lambda d: d.values(),
        ):
            for m, d in zip(f(self.model), f(self.dict)):
                assert m == d

    @rule(k=keys)
    def contains(self, k):
        assert (k in self.model) == (k in self.dict)

    @rule(other=st.dictionaries(keys, values))
    def update(self, other):
        self.model.update(other)
        self.dict.update(other)

    @rule(other=st.dictionaries(keys, values))
    def ior(self, other):
        self.model |= other
        self.dict |= other

    @rule(other=st.dictionaries(keys, values))
    def union(self, other):
        assert (self.model | other) == (self.dict | other)

    @rule()
    def clear(self):
        self.model.clear()
        self.dict.clear()

DictTest = DictComparison.TestCase

@settings(max_examples=25)
class ListComparison(RuleBasedStateMachine):
    indices = Bundle('indices')
    values = Bundle('values')
    heap = Heap()

    def __init__(self):
        super().__init__()
        self.list = Box[List](self.heap)
        self.model = []
        self.next_index = 0

    @rule(target=indices)
    def new_index(self):
        next_index = self.next_index
        self.next_index = next_index + 1
        return next_index

    @rule(target=values, v=st.text())
    def value(self, v):
        return v

    @rule()
    def len(self):
        assert len(self.model) == len(self.list)

    @rule(i=indices)
    def getitem(self, i):
        try:
            model_i = self.model[i]
        except IndexError:
            with pytest.raises(IndexError):
                self.list[i]
        else:
            assert model_i == self.list[i]

    @rule(i=indices, v=values)
    def setitem(self, i, v):
        try:
            self.model[i] = v
        except IndexError:
            with pytest.raises(IndexError):
                self.list[i] = v
        else:
            self.list[i] = v

    @rule(i=indices)
    def delitem(self, i):
        try:
            del self.model[i]
        except IndexError:
            with pytest.raises(IndexError):
                del self.list[i]
        else:
            del self.list[i]

    @rule(v=values, start=indices, stop=indices)
    def index(self, v, start, stop):
        try:
            model_i = self.model.index(v, start, stop)
        except ValueError:
            with pytest.raises(ValueError):
                self.list.index(v, start, stop)
        else:
            assert model_i == self.list.index(v, start, stop)

    @rule(v=values)
    def count(self, v):
        assert self.model.count(v) == self.list.count(v)

    @rule()
    def iters(self):
        for f in (
            iter,
            reversed
        ):
            for m, d in zip(f(self.model), f(self.list)):
                assert m == d

    @rule(v=values)
    def contains(self, v):
        assert (v in self.model) == (v in self.list)

    def extend_indices(self, l):
        start = self.next_index
        self.next_index += len(l)
        return multiple(*range(start, self.next_index))

    @rule(target=indices, other=st.lists(values))
    def iadd(self, other):
        self.model += other
        self.list += other
        return self.extend_indices(other)

    @rule(target=indices, i=indices, v=values)
    def insert(self, i, v):
        self.model.insert(i, v)
        self.list.insert(i, v)
        return self.new_index()

    @rule(target=indices, v=values)
    def append(self, v):
        self.model.append(v)
        self.list.append(v)
        return self.new_index()

    @rule()
    def reverse(self):
        self.model.reverse()
        self.list.reverse()

    @rule(target=indices, l=st.lists(values))
    def extend(self, l):
        self.model.extend(l)
        self.list.extend(l)
        return self.extend_indices(l)

    @rule(i=st.one_of(st.just(-1), indices))
    def pop(self, i):
        try:
            model_pop = self.model.pop(i)
        except IndexError:
            with pytest.raises(IndexError):
                self.list.pop(i)
        else:
            assert model_pop == self.list.pop(i)

    @rule(v=values)
    def remove(self, v):
        try:
            self.model.remove(v)
        except ValueError:
            with pytest.raises(ValueError):
                self.list.remove(v)
        else:
            self.list.remove(v)

ListTest = ListComparison.TestCase



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>
# Copyright (c) 2017 Viktor Adam
# Portions of this file are adapted from prometheus_flask_exporter

import time
import random
import multiprocessing
import requests

endpoints = ("one", "two", "three", "four", "error")
HOST = "http://localhost:8000/"

def run():
    while True:
        try:
            target = random.choice(endpoints)
            requests.get(HOST + target, timeout=1)
        except requests.RequestException:
            print("cannot connect", HOST)
            time.sleep(1)

if __name__ == "__main__":
    for _ in range(4):
        thread = multiprocessing.Process(target=run)
        thread.start()
    run()



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>
# Copyright 2015 The Prometheus Authors
# Portions of this file are adapted from prometheus_client

from mpmetrics import Summary
from prometheus_client import start_http_server
import multiprocessing
import random
import time

# Create a metric to track time spent and requests made.
REQUEST_TIME = Summary('request_processing_seconds', 'Time spent processing request')

# Decorate function with metric.
@REQUEST_TIME.time()
def process_request(t):
    """A dummy function that takes some time."""
    time.sleep(t)

# Function to generate requests
def generate_requests():
    while True:
        process_request(random.random())

if __name__ == '__main__':
    # Start up the server to expose the metrics.
    start_http_server(8000)
    # Generate some requests from two processes
    multiprocessing.Process(target=generate_requests).start()
    generate_requests()



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

"""multiprocess-safe metrics

mpmetrics implements metrics suitable for use with OpenMetrics. It provides
multiprocess-safe replacements for prometheus_client's Counter, Gauge, Summary,
Histogram, and Enum. To use it, just import these classes from this module
instead of from prometheus_client.
"""

from .metrics import Counter, Gauge, Summary, Histogram, Enum

__all__ = ('Counter', 'Gauge', 'Summary', 'Histogram', 'Enum')



import os
import sys
sys.path.insert(0, os.path.abspath('..'))

project = 'mpmetrics'
copyright = '2022-23 Sean Anderson'
author = 'Sean Anderson'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
]

exclude_patterns = ['out']
html_theme = 'sphinx_rtd_theme'

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
}



# SPDX-License-Identifier: GPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

from contextlib import contextmanager
import datetime
import time

from freezegun import freeze_time
from freezegun.api import fake_monotonic
time.perf_counter = fake_monotonic

import pytest
from prometheus_client.parser import text_string_to_metric_families

import examples.flask_server
from examples.flask_server import app
app.config['TESTING'] = True

@contextmanager
def sleeping_freeze_time(*args, **kwargs):
    with freeze_time(*args, **kwargs) as frozen:
        real_sleep = time.sleep
        time.sleep = frozen.tick
        yield frozen
        time.sleep = real_sleep

def test_flask_server():
    client = app.test_client()
    with sleeping_freeze_time():
        assert client.get('/one').status_code == 200
        assert client.get('/two').status_code == 200
        assert client.get('/three').status_code == 200
        assert client.get('/four').status_code == 200
        assert client.get('/error').status_code == 500

    resp = client.get('/metrics')
    assert resp.status_code == 200

    name = 'flask_http_request_duration_seconds'
    for metric in text_string_to_metric_families(resp.data.decode('utf-8')):
        if metric.name == name:
            break

    def get_sample(name, endpoint, le, method, status):
        for sample in metric.samples:
            if sample.name == name \
               and sample.labels['endpoint'] == endpoint \
               and sample.labels['le'] == le \
               and sample.labels['method'] == method \
               and sample.labels['status'] == status:
                return sample
        raise KeyError

    assert get_sample(name + '_bucket', 'first_route', '0.25', 'GET', '200').value == 1.0
    assert get_sample(name + '_bucket', 'the_second', '0.5', 'GET', '200').value == 1.0
    assert get_sample(name + '_bucket', 'test_3rd', '0.75', 'GET', '200').value == 1.0
    assert get_sample(name + '_bucket', 'fourth_one', '1.0', 'GET', '200').value == 1.0
    assert get_sample(name + '_bucket', 'oops', '0.005', 'GET', '500').value == 1.0



# SPDX-License-Identifier: GPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

import math
import mmap

import pytest
from hypothesis import assume, given, strategies as st

from mpmetrics.util import align, align_down, genmask

@given(st.integers(), st.integers(0, 100).map(lambda n: 1 << n))
def test_align(x, a):
    res = align(x, a)
    assert res % a == 0
    assert res >= x
    assert res - a < x

    res = align_down(x, a)
    assert res % a == 0
    assert res <= x
    assert res + a >= x

@given(st.integers(), st.integers(1).filter(lambda a: math.log2(a) % 1))
def test_npot(x, a):
    with pytest.raises(ValueError):
        align(x, a)

    with pytest.raises(ValueError):
        align_down(x, a)

@given(st.integers(0, 63), st.integers(0, 63))
def test_genmask(hi, lo):
    assume(hi >= lo)
    for bit in range(lo, hi):
        assert (1 << bit) & genmask(hi, lo)



# SPDX-License-Identifier: GPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

import math
import mmap

import pytest
from hypothesis import given, HealthCheck, settings, strategies as st

from mpmetrics.heap import Heap
from mpmetrics.util import align

from .common import parallel

@given(st.integers().filter(lambda n: n % mmap.PAGESIZE))
def test_bad_map_size(map_size):
    with pytest.raises(ValueError):
        Heap(map_size=map_size)

@given(st.integers(max_value=0))
def test_small_size(size):
    with pytest.raises(ValueError):
        Heap().malloc(size)

@given(st.integers().filter(lambda a: a <= 0 or math.log2(a) % 1))
def test_bad_align(alignment):
    with pytest.raises(ValueError):
        Heap().malloc(1, alignment)

@st.composite
def allocs(draw):
    size = mmap.ALLOCATIONGRANULARITY << draw(st.integers(0, 4))
    heap = Heap(map_size=size)
    sizes = st.integers(1, 2 * size)
    aligns = st.integers(0, 12).map(lambda n: 1 << n)
    return heap, draw(st.lists(st.tuples(sizes, aligns), min_size=3))

@given(allocs())
@settings(max_examples=25, suppress_health_check=(HealthCheck.data_too_large,))
def test_malloc(alloc):
    h, paramlist = alloc

    blocks = [h.malloc(*params) for params in paramlist]
    for ((size, alignment), block) in zip(paramlist, blocks):
        if size < h.map_size:
            assert block.size == size
        else:
            assert block.size >= size
        assert block.start == align(block.start, alignment)

    mems = [block.deref() for block in blocks]
    for (block, mem) in zip(blocks, mems):
        if size < h.map_size:
            assert len(mem) == block.size
        else:
            assert len(mem) >= block.size
        assert not any(mem)
        mem[:] = b'A' * len(mem)

    blocks.sort(key=lambda block: block.start)
    for i in range(len(blocks)):
        if i:
            prev = blocks[i - 1]
            assert prev.start + prev.size <= blocks[i].start

def set_pre(block, val):
    block.deref()[0] = val

def test_prefork(parallel):
    block = Heap().malloc(1)
    mem = block.deref()
    assert mem[0] == 0

    p = parallel.spawn(target=set_pre, args=(block, 1))
    p.start()
    p.join()
    assert mem[0] == 1

def set_post(q, val):
    mem = q.get().deref()
    mem[0] = val

def test_postfork(parallel):
    q1 = parallel.queue()
    p1 = parallel.spawn(target=set_post, args=(q1, 1))
    p1.start()

    q2 = parallel.queue()
    p2 = parallel.spawn(target=set_post, args=(q2, 2))
    p2.start()

    h = Heap()
    h.malloc(mmap.PAGESIZE)
    block = h.malloc(1)
    mem = block.deref()
    assert mem[0] == 0
    q1.put(block)
    p1.join()
    assert mem[0] == 1

    q2.put(block)
    p2.join()
    assert mem[0] == 2

def chain_start(q1, q2):
    h = Heap()
    h.malloc(mmap.PAGESIZE)
    block = h.malloc(1)
    q1.put(block)
    q2.get()

class DummyBlock:
    def deref(self):
        return [0]

def test_chain(parallel):
    q1 = parallel.queue()
    q2 = parallel.queue()
    p1 = parallel.spawn(target=chain_start, args=(q1, q2))
    p1.start()

    q3 = parallel.queue()
    p2 = parallel.spawn(target=set_post, args=(q3, 2))
    p2.start()

    block = DummyBlock()
    try:
        block = q1.get()
        q2.put(None)
        mem = block.deref()
        assert mem[0] == 0
        mem[0] = 1
        p1.join()
    finally:
        q3.put(block)
        p2.join()
    assert mem[0] == 2



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

"""Various types backed by (shared) memory"""

import ctypes
import io
from multiprocessing.reduction import ForkingPickler
import pickle
import sys

from .generics import IntType, ObjectType, ProductType
from .util import align, classproperty

def _wrap_ctype(__name__, ctype, doc):
    size = ctypes.sizeof(ctype)
    align = ctypes.alignment(ctype)

    __doc__ = f"""{doc.capitalize()} backed by (shared) memory.

    .. py:attribute:: value
        :type: {'float' if __name__ == 'Double' else 'int'}

        The value itself. You can read and modify this value as necessary. For
        example::

            from mpmetrics.heap import Heap
            from mpmetrics.types import Box, {__name__}

            var = Box[{__name__}](Heap())
            assert var.value == 0
            var.value += 1
            assert var.value == 1

    .. py:attribute:: size
        :type: int
        :value: {size}

        The size of {doc}, in bytes

    .. py:attribute:: align
        :type: int
        :value: {align}

        The alignment of {doc}, in bytes
    """

    def __init__(self, mem, heap=None):
        self._mem = mem
        self._value = ctype.from_buffer(mem)
        self._value.value = 0

    __init__.__doc__ = f"""Create a new {__name__}.

    :param memoryview mem: The backing memory
    :param heap: Unused
    """

    def _setstate(self, mem, heap=None, **kwargs):
        self._mem = mem
        self._value = ctype.from_buffer(mem)

    def __getattr__(self, name):
        return getattr(self.__dict__['_value'], name)

    def __setattr__(self, name, value):
        if '_value' in self.__dict__:
            setattr(self.__dict__['_value'], name, value)
        else:
            self.__dict__[name] = value

    def __delattr__(self, name):
        delattr(self.__dict__['_value'], name)

    ns = locals()
    del ns['doc']
    del ns['ctype']
    return type(__name__, (), ns)

Double = _wrap_ctype('Double', ctypes.c_double, "a double")
Size_t = _wrap_ctype('Size_t', ctypes.c_size_t, "a size_t")
Int64 = _wrap_ctype('Int64', ctypes.c_int64, "an int64_t")
UInt64 = _wrap_ctype('UInt64', ctypes.c_uint64, "a uint64_t")

class Struct:
    """A structured group of fields backed by (shared) memory.

    This is a base class that can be subclassed to create C-style structs::

        from mpmetrics.heap import Heap
        from mpmetrics.types import Double, Size_t, Struct

        class MyStruct(mpmetrics.types.Struct):
            _fields_ = {
                'a': Double,
                'b': Size_t,
            }

        assert MyStruct.size == Double.size + Size_t.size
        s = Box[MyStruct](Heap())
        assert type(s.a) == Double
        assert type(s.b) == Size_t

    .. py:property:: _fields_
        :classmethod:
        :type: dict[str, Any]

        The fields of the struct, in order. Upon initialization, each value is
        initialized with a block of memory equal to its `.size`. Padding is
        added as necessary to ensure alignment.

        Subclasses must implement this property.
    """

    @classmethod
    def _fields_iter(cls):
        off = 0
        for name, field in cls._fields_.items():
            off = align(off, field.align)
            yield name, field, off
            off += field.size

    @classproperty
    def size(cls):
        """The size of the struct, in bytes"""
        for name, field, off in cls._fields_iter():
            size = field.size + off
        return size

    @classproperty
    def align(cls):
        """The alignment of the struct, in bytes"""
        return max(field.align for field in cls._fields_.values())

    def __init__(self, mem, heap=None):
        """Create a new Struct.

        :param memoryview mem: The backing memory
        :param mpmetrics.heap.Heap heap: Passed to each field's ``__init__``
        """

        self._mem = mem
        for name, field, off in self._fields_iter():
            setattr(self, name, field(mem[off:off + field.size], heap=heap))

    def _setstate(self, mem, heap=None, **kwargs):
        self._mem = mem
        for name, field, off in self._fields_iter():
            field = field.__new__(field)
            field._setstate(self._mem[off:off + field.size], heap=heap)
            setattr(self, name, field)

def Array(__name__, cls, n):
    """An array of values backed by (shared) memory.

    You can access values in an `Array` just like it was a `list`::

        from mpmetrics.heap import Heap
        from mpmetrics.types import Array, Box, Double

        assert Array[Double, 5].size == Double.size * 5
        a = Box[Array[Double, 5]](Heap())
        assert type(a[0]) == Double
        a[4].value = 6.28


    .. py:method:: Array.__init__(mem, heap=None)

        Create a new Array.

        :param memoryview mem: The backing memory
        :param mpmetrics.heap.Heap heap: Passed to each member's ``__init__``
    """

    if n < 1:
        raise ValueError("n must be strictly positive")

    member_size = align(cls.size, cls.align)
    size = member_size * n

    def __init__(self, mem, heap=None):
        self._mem = mem
        self._vals = []
        for i in range(n):
            off = i * member_size
            self._vals.append(cls(self._mem[off:off + member_size], heap=heap))

    def _setstate(self, mem, heap=None, **kwargs):
        self._mem = mem
        self._vals = []
        for i in range(n):
            off = i * member_size
            val = cls.__new__(cls)
            val._setstate(self._mem[off:off + member_size], heap=heap)
            self._vals.append(val)

    def __len__(self):
        return n

    def __getitem__(self, key):
        return self._vals[key]

    def __iter__(self):
        return iter(self._vals)

    ns = locals()
    ns['align'] = cls.align
    del ns['member_size']
    del ns['cls']
    del ns['n']

    return type(__name__, (), ns)

Array = ProductType('Array', Array, (ObjectType, IntType))

class _Box:
    """A heap-allocated box to put values in

    This class "boxes" another class using heap-allocated memory. For example,
    you could create a `Double` like::

        from mpmetrics.heap import Heap
        from mpmetrics.types import Double

        block = Heap().malloc(Double.size)
        d = Double(block.deref())

    But d._mem is a `memoryview` which can't be pickled. `Box` takes care of
    keeping track of the memory block::

        from mpmetrics.heap import Heap
        from mpmetrics.types import Box, Double

        d = Box[Double](Heap())

    .. py:method:: Box.__init__(heap, *args, **kwargs)

        Create a new object on the heap

        :param mpmetrics.heap.Heap heap: The heap to use when allocating the object
        :param \\*args: Any additional arguments are passed to the boxed class
        :param \\**kwargs: Any additional keyword arguments are passed to the boxed class.

        The superclass's `__init__` is called with a newly-allocated buffer as
        the first argument, any positional arguments to this function, the
        keyword argument `heap` set to `heap`, and any additional keyword
        arguments to this function.

    .. py:method:: Box._getstate()
        :abstractmethod:

        Return keyword arguments to pass to `_setstate`.

        :return: A dictionary of keyword arguments for `_setstate`
        :rtype: dict

        This method is optional; if it is not implemented then no additional
        keyword arguments will be passed to `_setstate`.

    .. py:method:: Box._setstate(mem, heap=None, **kwargs)
        :abstractmethod:

        Initialize internal state after unpickling.

        :param memoryview mem: The backing memory
        :param mpmetrics.heap.Heap heap: The heap `mem` was allocated from
        :param \**kwargs: Any additional arguments from `_getstate`

        This method must be implemented by boxed types.
    """

    def __init__(self, heap, *args, **kwargs):
        block = heap.malloc(self.size)
        super().__init__(block.deref(), *args, heap=heap, **kwargs)
        self.__block = block

    def __getstate__(self):
        try:
            kwargs = super()._getstate()
        except AttributeError:
            kwargs = {}
        return self.__block, kwargs

    def __setstate__(self, state):
        self.__block, kwargs = state
        super()._setstate(self.__block.deref(), heap=self.__block.heap, **kwargs)

def _create_box(name, cls):
    return type(name, (_Box, cls), {'__doc__': cls.__doc__})

_create_box.__doc__ = _Box.__doc__
Box = ObjectType('Box', _create_box)

class _Pickler(ForkingPickler):
    def __init__(self, file, heap, protocol=None):
        super().__init__(file, protocol)
        self.heap = heap

    def persistent_id(self, obj):
        if obj is self.heap:
            return 'heap'

    @classmethod
    def dumps(cls, obj, heap, protocol=None):
        buf = io.BytesIO()
        cls(buf, heap, protocol).dump(obj)
        return buf.getbuffer()

class _Unpickler(pickle.Unpickler):
    def __init__(self, file, heap):
        super().__init__(file)
        self.heap = heap

    def persistent_load(self, pid):
        if pid != 'heap':
            raise pickle.UnpicklingError(f"unsupported persistent object {pid}")
        return self.heap

    @classmethod
    def loads(cls, data, heap):
        buf = io.BytesIO(data)
        return cls(buf, heap).load()

class Object(Struct):
    """A python object pickled in (shared) memory

    This is a base class for python objects backed by shared memory. Whenever
    the object is accessed, it is unpickled from the backing memory. When it is
    modified, it is pickled to the backing memory.

    This class itself does not contain the actual object. Instead, it contains
    the start/size/length of the block containing the object. When the object
    grows too large for the block, the old block is free'd and a new one is
    allocated.

    This class provides no synchronization. All methods should be accessed
    under some other form of synchonization, such as a
    :py:class:`_mpmetrics.Lock`.
    """

    _fields_ = {
        '_start': Size_t,
        '_size': Size_t,
        '_len': Size_t,
    }

    def __init__(self, mem, heap):
        """Create a new Object.

        :param memoryview mem: The memory used to store information about the buffer
        :param mpmetrics.heap.Heap heap: The heap to use when (re)allocating the buffer
        """

        if not heap:
            raise ValueError("heap must be provided")
        super().__init__(mem)
        self._heap = heap

    def _setstate(self, mem, heap):
        assert heap is not None
        super()._setstate(mem, heap)
        self._heap = heap

    @property
    def _block(self):
        if self._size.value:
            return self._heap.Block(self._heap, self._start.value, self._size.value)

    @_block.setter
    def _block(self, block):
        self._start.value = block.start
        self._size.value = block.size

    @property
    def _object(self):
        if self._len.value:
            return _Unpickler.loads(self._block.deref()[:self._len.value], self._heap)
        return self._new()

    @_object.setter
    def _object(self, v):
        vs = _Pickler.dumps(v, self._heap)
        new_length = len(vs)
        self._len.value = new_length
        if self._len.value > self._size.value:
            if self._size.value:
                self._block.free()
            # Scale by a lot to minimize allocations; Heap doesn't free backing memory
            self._block = self._heap.malloc(4 * new_length)
        self._block.deref()[:self._len.value] = vs

    def _mutate(self, method, *args, **kwargs):
        v = self._object
        result = getattr(v, method)(*args, **kwargs)
        self._object = v
        return result

    def __repr__(self):
        return f"{self.__class__.__qualname__}({repr(self._object)})"

class Sized:
    def __len__(self):
        return len(self._object)

class Iterable:
    def __iter__(self):
        return iter(self._object)

class Container:
    def __contains__(self, item):
        return item in self._object

class Collection(Sized, Iterable, Container):
    pass

class Reversible:
    def __reversed__(self):
        return reversed(self._object)

class Sequence(Reversible, Collection):
    def __len__(self):
        return len(self._object)

    def __getitem__(self, key):
        return self._object[key]

    def index(self, value, start=0, stop=sys.maxsize):
        return self._object.index(value, start, stop)

    def count(self, value):
        return self._object.count(value)

class MutableSequence(Sequence):
    def __setitem__(self, key, value):
        v = self._object
        v[key] = value
        self._object = v

    def __delitem__(self, key):
        v = self._object
        del v[key]
        self._object = v

    def __iadd__(self, other):
        self._object = self._object + other
        return self._object

    def insert(self, index, object):
        self._mutate('insert', index, object)

    def append(self, object):
        self._mutate('append', object)

    def reverse(self):
        self._mutate('reverse')

    def extend(self, iterable):
        self._mutate('extend', iterable)

    def pop(self, index=-1):
        return self._mutate('pop', index)

    def remove(self, value):
        self._mutate('remove', value)

class Mapping(Collection):
    def __getitem__(self, key):
        return self._object[key]

    def __eq__(self, other):
        return self._object == other

    def __neq__(self, other):
        return self._object != other

    def get(self, key, default=None):
        return self._object.get(key, default)

    def items(self):
        return self._object.items()

    def keys(self):
        return self._object.keys()

    def values(self):
        return self._object.values()

class MutableMapping(Mapping):
    def __setitem__(self, key, value):
        v = self._object
        v[key] = value
        self._object = v

    def __delitem__(self, key):
        v = self._object
        del v[key]
        self._object = v

    def clear(self):
        self._len.value = 0

    def pop(self, key, default=None):
        return self._mutate('pop', key, default)

    def popitem(self):
        return self._mutate('popitem')

    def setdefault(self, key, default=None):
        return self._mutate('setdefault', key, default)

    def update(self, other=()):
        return self._mutate('update', other)

class Dict(Object, Sequence, MutableMapping):
    """A `dict` backed by (shared) memory.

    All methods of `dict` are supported. External synchronization (such as from
    a :py:class:`_mpmetrics.Lock`) must be provided when accessing any method.
    """

    _new = dict

    def __or__(self, other):
        return self._object | other

    def __ior__(self, other):
        self._object = self._object | other
        return self

    def copy(self):
        return self._object

class List(Object, MutableSequence):
    """A `list` backed by (shared) memory.

    All methods of `list` are supported. External synchronization (such as from
    a :py:class:`_mpmetrics.Lock`) must be provided when accessing any method.
    """

    _new = list

    def sort(self, key=None, reverse=False):
        self._mutate('sort', key, reverse)



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

"""Import this module to monkey-patch `prometheus_flask_exporter` to use mpmetrics' metrics."""

import prometheus_flask_exporter

from . import Counter, Gauge, Summary, Histogram, Enum

prometheus_flask_exporter.Gauge = Gauge
prometheus_flask_exporter.Counter = Counter
prometheus_flask_exporter.Summary = Summary
prometheus_flask_exporter.Histogram = Histogram
prometheus_flask_exporter.Enum = Enum

del Gauge
del Counter
del Summary
del Histogram
del Enum

del prometheus_flask_exporter

from prometheus_flask_exporter import *



# SPDX-License-Identifier: GPL-3.0-only AND Apache-2.0
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>
# Copyright 2015 The Prometheus Authors
# Portions of this file are adapted from prometheus_client

from contextlib import nullcontext
import pickle
import random
import time

from hypothesis import given, strategies as st
from prometheus_client.registry import CollectorRegistry
import pytest

from mpmetrics.metrics import Counter, Enum, Gauge, Summary, Histogram, _validate_exemplar
from mpmetrics.atomic import AtomicUInt64

from .common import heap, parallel, parallels, ParallelLoop

def test_exemplar_validation():
    # 128 characters should not raise, even using characters larger than 1 byte.
    _validate_exemplar({
        'abcdefghijklmnopqrstuvwxyz': '26+16 characters',
        'x123456': '7+15 characters',
        'zyxwvutsrqponmlkjihgfedcba': '26+16 characters',
        'unicode': '7+15 chars    平',
    })

    # 129 characters in total should fail.
    with pytest.raises(ValueError):
        _validate_exemplar({
            'abcdefghijklmnopqrstuvwxyz': '26+16 characters',
            'x1234567': '8+15 characters',
            'zyxwvutsrqponmlkjihgfedcba': '26+16 characters',
            'y123456': '7+15 characters',
        })

    with pytest.raises(ValueError):
        _validate_exemplar({':o)': 'smile'})
    with pytest.raises(ValueError):
        _validate_exemplar({'1': 'number'})

@pytest.fixture(scope='session')
def registry(heap):
    class FakeRegistry:
        def register(self, collector):
            pass

    return FakeRegistry()

def get_sample_value(collector, name, labels={}):
    for metric in collector.collect():
        for s in metric.samples:
            if s.name == name and s.labels == labels:
                return s.value
    return None

def get_sample_exemplar(collector, name, labels={}):
    for metric in collector.collect():
        for s in metric.samples:
            if s.name == name and s.labels == labels:
                return s.exemplar
    return None

class TestCounter:
    @pytest.fixture
    def counter(self, registry):
        return Counter('c_total', "help", registry=registry)

    def test_increment(self, counter):
        assert get_sample_value(counter, 'c_created')
        assert get_sample_value(counter, 'c_total') == 0
        assert get_sample_exemplar(counter, 'c_total') is None
        counter.inc()
        assert get_sample_value(counter, 'c_total') == 1
        counter.inc(7, {'foo': 'bar'})
        assert get_sample_value(counter, 'c_total') == 8
        exemplar = get_sample_exemplar(counter, 'c_total')
        assert exemplar.labels == {'foo': 'bar'}
        assert exemplar.value == 7
        assert exemplar.timestamp

        with pytest.raises(OverflowError):
            counter.inc(AtomicUInt64.max)

    @given(st.integers(max_value=-1))
    def test_negative_increment_raises(self, registry, amount):
        counter = Counter('c_total', "help", registry=registry)
        with pytest.raises(ValueError):
            counter.inc(amount)

    def test_exceptions(self, counter):
        for exception in (ValueError, TypeError):
            try:
                with counter.count_exceptions(ValueError):
                    raise exception
            except:
                pass

        assert get_sample_value(counter, 'c_total') == 1

        for x in range(2):
            try:
                counter.count_exceptions(ZeroDivisionError)(lambda: 1/x)()
            except:
                pass

        assert get_sample_value(counter, 'c_total') == 2

    class ConcurrentTest(ParallelLoop):
        def __init__(self, counter, parallel):
            super().__init__(parallel)
            self.counter = counter

        def loop(self, n):
            self.counter.inc()

        def final(self):
            assert get_sample_value(self.counter, 'c_total') == self.total

    def test_concurrent(self, counter, parallel):
        self.ConcurrentTest(counter, parallel).run()

class TestGauge:
    @pytest.fixture
    def gauge(self, registry):
        return Gauge('g', 'help', registry=registry)

    def test_gauge(self, gauge):
        assert get_sample_value(gauge, 'g') == 0
        gauge.inc()
        assert get_sample_value(gauge, 'g') == 1
        gauge.dec(3)
        assert get_sample_value(gauge, 'g') == -2
        gauge.set(9)
        assert get_sample_value(gauge, 'g') == 9

    def test_inprogress(self, gauge):
        @gauge.track_inprogress()
        def test():
            assert get_sample_value(gauge, 'g') == 1
            with gauge.track_inprogress():
                assert get_sample_value(gauge, 'g') == 2
                raise Exception
            assert get_sample_value(gauge, 'g') == 1

        try:
            test()
        except:
            pass
        assert get_sample_value(gauge, 'g') == 0

    class ConcurrentTest(ParallelLoop):
        def __init__(self, gauge, parallel):
            super().__init__(parallel)
            self.gauge = gauge

        def loop(self, n):
            self.gauge.inc(1)

        def final(self):
            assert get_sample_value(self.gauge, 'g') == self.total

    def test_concurrent(self, gauge, parallel):
        self.ConcurrentTest(gauge, parallel).run()

class TestSummary:
    @pytest.fixture
    def summary(self, registry):
        return Summary('s', 'help', registry=registry)

    def test_summary(self, summary):
        assert get_sample_value(summary, 's_created')
        assert get_sample_value(summary, 's_count') == 0
        assert get_sample_value(summary, 's_sum') == 0
        summary.observe(10)
        assert get_sample_value(summary, 's_count') == 1
        assert get_sample_value(summary, 's_sum') == 10

    class ConcurrentTest(ParallelLoop):
        def __init__(self, summary, parallel):
            super().__init__(parallel)
            self.summary = summary

        def get_samples(self):
            metric = next(self.summary.collect())
            for s in metric.samples:
                if s.name.endswith('s_count'):
                    count = s.value
                elif s.name.endswith('s_sum'):
                    sum = s.value
            return count, sum

        def loop(self, n):
            self.summary.observe(1)

        def check(self):
            count, sum = self.get_samples()
            assert count == sum

        def final(self):
            count, sum = self.get_samples()
            assert count == self.total
            assert sum == self.total

    def test_concurrent(self, summary, parallel):
        self.ConcurrentTest(summary, parallel).run()

class TestHistogram:
    @pytest.fixture
    def histogram(self, registry):
        return Histogram('h', 'help', registry=registry)

    @pytest.fixture
    def labels(self, registry):
        return Histogram('h', 'help', ['l'], registry=registry)

    def test_histogram(self, histogram):
        assert get_sample_value(histogram, 'h_created')
        assert get_sample_value(histogram, 'h_bucket', {'le': '1.0'}) == 0
        assert get_sample_value(histogram, 'h_bucket', {'le': '2.5'}) == 0
        assert get_sample_value(histogram, 'h_bucket', {'le': '5.0'}) == 0
        assert get_sample_value(histogram, 'h_bucket', {'le': 'inf'}) == 0
        assert get_sample_value(histogram, 'h_count') == 0
        assert get_sample_value(histogram, 'h_sum') == 0
        for le in ('1.0', '2.5', '5.0', 'inf'):
            assert get_sample_exemplar(histogram, 'h_bucket', {'le': le}) is None

        histogram.observe(2, {'foo': 'bar'})
        assert get_sample_value(histogram, 'h_bucket', {'le': '1.0'}) == 0
        assert get_sample_value(histogram, 'h_bucket', {'le': '2.5'}) == 1
        assert get_sample_value(histogram, 'h_bucket', {'le': '5.0'}) == 1
        assert get_sample_value(histogram, 'h_bucket', {'le': 'inf'}) == 1
        assert get_sample_value(histogram, 'h_count') == 1
        assert get_sample_value(histogram, 'h_sum') == 2
        exemplar = get_sample_exemplar(histogram, 'h_bucket', {'le': '2.5'})
        assert exemplar.labels == {'foo': 'bar'}
        assert exemplar.value == 2
        assert exemplar.timestamp

        histogram.observe(2.5, {'foo': 'baz'})
        assert get_sample_value(histogram, 'h_bucket', {'le': '1.0'}) == 0
        assert get_sample_value(histogram, 'h_bucket', {'le': '2.5'}) == 2
        assert get_sample_value(histogram, 'h_bucket', {'le': '5.0'}) == 2
        assert get_sample_value(histogram, 'h_bucket', {'le': 'inf'}) == 2
        assert get_sample_value(histogram, 'h_count') == 2
        assert get_sample_value(histogram, 'h_sum') == 4.5
        assert get_sample_exemplar(histogram, 'h_bucket', {'le': '2.5'}).labels['foo'] == 'baz'

        histogram.observe(float("inf"))
        assert get_sample_value(histogram, 'h_bucket', {'le': '1.0'}) == 0
        assert get_sample_value(histogram, 'h_bucket', {'le': '2.5'}) == 2
        assert get_sample_value(histogram, 'h_bucket', {'le': '5.0'}) == 2
        assert get_sample_value(histogram, 'h_bucket', {'le': 'inf'}) == 3
        assert get_sample_value(histogram, 'h_count') == 3
        assert get_sample_value(histogram, 'h_sum') == float("inf")

    def test_setting_buckets(self, registry):
        def get_buckets(h):
            buckets = []
            for sample in next(h.collect()).samples:
                if sample.name != 'h_bucket':
                    continue
                buckets.append(float(sample.labels['le']))
            return buckets

        h = Histogram('h', 'help', registry=registry, buckets=[0, 1, 2])
        assert get_buckets(h) == [0.0, 1.0, 2.0, float("inf")]

        h = Histogram('h', 'help', registry=registry, buckets=[0, 1, 2, float("inf")])
        assert get_buckets(h) == [0.0, 1.0, 2.0, float("inf")]

        with pytest.raises(ValueError):
            Histogram('h', 'help', registry=registry, buckets=[])

        with pytest.raises(ValueError):
            Histogram('h', 'help', registry=registry, buckets=[float('inf')])

        with pytest.raises(ValueError):
            Histogram('h', 'help', registry=registry, buckets=[3, 1])

    def test_invalid_labels(self, registry):
        with pytest.raises(ValueError):
            Histogram('h', 'help', registry=registry, labelnames=['le'])

    def test_labels(self, labels):
        labels.labels('a').observe(2)
        assert get_sample_value(labels, 'h_bucket', {'le': '1.0', 'l': 'a'}) == 0
        assert get_sample_value(labels, 'h_bucket', {'le': '2.5', 'l': 'a'}) == 1
        assert get_sample_value(labels, 'h_bucket', {'le': '5.0', 'l': 'a'}) == 1
        assert get_sample_value(labels, 'h_bucket', {'le': 'inf', 'l': 'a'}) == 1
        assert get_sample_value(labels, 'h_count', {'l': 'a'}) == 1
        assert get_sample_value(labels, 'h_sum', {'l': 'a'}) in (2, None)

    class ConcurrentTest(ParallelLoop):
        def __init__(self, histogram, parallel):
            super().__init__(parallel)
            self.histogram = histogram

        def get_sample(self):
            metric = next(self.histogram.collect())
            buckets = []
            for s in metric.samples:
                if s.name.endswith('h_bucket'):
                    buckets.append(s.value)
                elif s.name.endswith('h_sum'):
                    sum = s.value
                elif s.name.endswith('h_count'):
                    count = s.value
            return buckets, sum, count

        def loop(self, n):
            self.histogram.observe(random.choice(self.histogram.thresholds[:-1]))

        def check(self):
            buckets, sum, count = self.get_sample()
            prev = 0
            bucket_sum = 0
            for bucket, threshold in zip(buckets, self.histogram.thresholds):
                if threshold == float('inf'):
                    break
                bucket_sum += (bucket - prev) * threshold
                prev = bucket
            assert sum == pytest.approx(bucket_sum)
            assert buckets[-1] == count

        def final(self):
            buckets, sum, count = self.get_sample()
            assert buckets[-1] == count == self.total

    def test_concurrent(self, histogram, parallel):
        self.ConcurrentTest(histogram, parallel).run()

@pytest.mark.parametrize(('cls', 'name'), (
    (Gauge, 'm'),
    (Summary, 'm_sum'),
    (Histogram, 'm_sum'))
)
def test_time(registry, cls, name):
    metric = cls('m', 'help', registry=registry)
    with metric.time():
        time.sleep(0.001)
    assert get_sample_value(metric, name) >= 0.001

    metric = cls('m', 'help', registry=registry)
    metric.time()(time.sleep)(0.001)
    assert get_sample_value(metric, name) >= 0.001

@pytest.mark.parametrize('cls', (Counter, Gauge, Summary, Histogram))
def test_pickle(registry, cls):
    metric = cls('name', 'help', labelnames=('l'), registry=registry)
    metric.labels('x')
    assert list(metric.collect()) == list(pickle.loads(pickle.dumps(metric)).collect())

class TestEnum:
    @pytest.fixture
    def enum(self, registry):
        return Enum('e', 'help', states=['a', 'b', 'c'], registry=registry)

    @pytest.fixture
    def labels(self, registry):
        return Enum('el', 'help', ['l'], states=['a', 'b', 'c'], registry=registry)

    def test_enum(self, enum, registry):
        assert get_sample_value(enum, 'e', {'e': 'a'}) == 1
        assert get_sample_value(enum, 'e', {'e': 'b'}) == 0
        assert get_sample_value(enum, 'e', {'e': 'c'}) == 0

        enum.state('b')
        assert get_sample_value(enum, 'e', {'e': 'a'}) == 0
        assert get_sample_value(enum, 'e', {'e': 'b'}) == 1
        assert get_sample_value(enum, 'e', {'e': 'c'}) == 0

        with pytest.raises(ValueError):
            enum.state('d')

        with pytest.raises(TypeError):
            Enum('e', 'help', registry=registry)

    def test_labels(self, labels):
        labels.labels('a').state('c')
        assert get_sample_value(labels, 'el', {'l': 'a', 'el': 'a'}) == 0
        assert get_sample_value(labels, 'el', {'l': 'a', 'el': 'b'}) == 0
        assert get_sample_value(labels, 'el', {'l': 'a', 'el': 'c'}) == 1

    def test_overlapping_labels(self, registry):
        with pytest.raises(ValueError):
            Enum('e', 'help', registry=registry, labelnames=['e'])

    class ConcurrentTest(ParallelLoop):
        def __init__(self, enum, parallel):
            super().__init__(parallel)
            self.enum = enum

        def loop(self, n):
            self.enum.state(['a', 'b', 'c'][n % 3])

        def final(self):
            for metric in self.enum.collect():
                assert sum(sample.value for sample in metric.samples) == 1

    def test_concurrent(self, enum, parallel):
        self.ConcurrentTest(enum, parallel).run()

class TestLabelCollector:
    @pytest.fixture
    def counter(self, registry):
        return Counter('c_total', "help", labelnames=['l'], registry=registry)

    @pytest.fixture
    def two_labels(self, registry):
        return Counter('two', "help", labelnames=['a', 'b'], registry=registry)

    def test_child(self, counter, two_labels):
        counter.labels('x').inc()
        assert get_sample_value(counter, 'c_total', {'l': 'x'}) == 1
        two_labels.labels('x', 'y').inc(2)
        assert get_sample_value(two_labels, 'two_total', {'a': 'x', 'b': 'y'}) == 2

    def test_incorrect_label_count_raises(self, counter):
        with pytest.raises(ValueError):
            counter.labels()

        with pytest.raises(ValueError):
            counter.labels('a', 'b')

    def test_labels_on_labels(self, counter):
        with pytest.raises(AttributeError):
            counter.labels('a').labels('b')

    def test_labels_coerced_to_string(self, counter):
        counter.labels(None).inc()
        counter.labels(l=None).inc()
        assert get_sample_value(counter, 'c_total', {'l': 'None'}) == 2

    def test_non_string_labels_raises(self, counter):
        class Test:
            __str__ = None

        with pytest.raises(TypeError):
            counter.labels(Test())

        with pytest.raises(TypeError):
            counter.labels(l=Test())

    def test_namespace_subsystem_concatenated(self, registry):
        c = Counter('c_total', 'help', namespace='a', subsystem='b', registry=registry)
        assert c._name == 'a_b_c'

    def test_labels_by_kwarg(self, counter, two_labels):
        counter.labels(l='x').inc()
        assert get_sample_value(counter, 'c_total', {'l': 'x'}) == 1
        for kwargs in ({'l': 'x', 'm': 'y'}, {'m': 'y'}, {}):
            with pytest.raises(ValueError):
                counter.labels(**kwargs)

        two_labels.labels(a='x', b='y').inc()
        assert get_sample_value(two_labels, 'two_total', {'a': 'x', 'b': 'y'}) == 1
        for kwargs in (
            { 'a': 'x', 'b': 'y', 'c': 'z' },
            { 'a': 'x', 'c': 'z' },
            { 'b': 'y', 'c': 'z' },
            { 'c': 'z' }):
            with pytest.raises(ValueError):
                counter.labels(**kwargs)

        with pytest.raises(ValueError):
            two_labels.labels({'a': 'x'}, b='y')

    def test_invalid_names_raise(self, registry):
        for kwargs in (
            { 'name': '' },
            { 'name': '^' },
            { 'name': '', 'namespace': '&' },
            { 'name': '', 'subsystem': '(' },
            { 'name': 'c_total', 'labelnames': ['^'] },
            { 'name': 'c_total', 'labelnames': ['a:b'] },
            { 'name': 'c_total', 'labelnames': ['__reserved'] }):
            with pytest.raises(ValueError):
                Counter(**kwargs, documentation='', registry=registry)

        with pytest.raises(ValueError):
            Summary('c_total', '', labelnames=['quantile'], registry=registry)

    def test_empty_labels_list(self, registry):
        h = Histogram('h', 'help', [], registry=registry)
        assert get_sample_value(h, 'h_count') == 0

    def test_unit_appended(self, registry):
        h = Histogram('h', 'help', [], registry=registry, unit="seconds")
        assert h._name == 'h_seconds'

    def test_unit_notappended(self, registry):
        h = Histogram('h_seconds', 'help', [], registry=registry, unit="seconds")
        assert h._name == 'h_seconds'

    def test_no_units_for_info_enum(self, registry):
        #with pytest.raises(ValueError):
        #    Info('foo', 'help', unit="x")

        with pytest.raises(ValueError):
            Enum('foo', 'help', unit="x")

    def test_name_cleanup_before_unit_append(self, registry):
        c = Counter('b_total', 'help', unit="total", labelnames=['l'], registry=registry)
        assert c._name == 'b_total'

    class ConcurrentTest(ParallelLoop):
        def __init__(self, counter, parallel):
            super().__init__(parallel)
            self.counter = counter

        def loop(self, n):
            self.counter.labels(n % self.n).inc()

        def final(self):
            for i in range(self.n):
                assert get_sample_value(self.counter, 'c_total', {'l': str(i)}) == self.count

    def test_concurrent(self, counter, parallel):
        self.ConcurrentTest(counter, parallel).run()
